"""Implementation of the main build function."""

import logging
import os
import re
import sys
from pathlib import Path
from subprocess import check_call, check_output

try:
    import tomllib  # type: ignore
except ModuleNotFoundError:
    import tomli as tomllib  # type: ignore

from .cmeel import __version__
from .config import cmeel_config
from .consts import CMEEL_PREFIX, SITELIB
from .metadata import metadata
from .utils import (
    deprecate_build_system,
    ensure_relocatable,
    expose_bin,
    get_tag,
    launch_tests,
    log_pip,
    normalize,
    patch,
)

LOG = logging.getLogger("cmeel.impl")


def build_impl(wheel_directory, editable=False) -> str:
    """Run CMake configure / build / test / install steps, and pack the wheel."""
    logging.basicConfig(level=cmeel_config.log_level.upper())
    LOG.info("CMake Wheel in editable mode" if editable else "CMake Wheel")
    LOG.info("cmeel version %s", __version__)
    log_pip()

    prefix = Path() / "build-editable" if editable else cmeel_config.temp_dir
    build = prefix / "bld"
    wheel_dir = prefix / "whl"
    install = (prefix if editable else wheel_dir) / CMEEL_PREFIX

    LOG.info("load conf from pyproject.toml")
    with Path("pyproject.toml").open("rb") as f:
        pyproject = tomllib.load(f)

    conf = pyproject["project"]
    conf["name"] = normalize(conf["name"])
    distribution = f"{conf['name'].replace('-', '_')}-{conf['version']}"

    source = deprecate_build_system(pyproject, "source", ".")
    run_tests = (
        os.environ.get("CMEEL_RUN_TESTS", "ON").upper()
        not in ("0", "NO", "OFF", "FALSE")
        if "CMEEL_RUN_TESTS" in os.environ
        else deprecate_build_system(pyproject, "run-tests", True)
    )
    run_tests_after_install = deprecate_build_system(
        pyproject,
        "run-tests-after-install",
        False,
    )
    build_number = deprecate_build_system(pyproject, "build-number", 0)
    configure_args = deprecate_build_system(pyproject, "configure-args", [])

    check_relocatable = deprecate_build_system(pyproject, "check-relocatable", True)
    fix_pkg_config = deprecate_build_system(pyproject, "fix-pkg-config", True)

    LOG.info("build wheel")

    # Patch

    patch()

    # Set env

    if run_tests_after_install:
        path = f"{install / SITELIB}"
        old = os.environ.get("PYTHONPATH", "")
        if old:
            path += f"{os.pathsep}{old}"
        os.environ.update(PYTHONPATH=path)

    # Configure

    LOG.info("configure")
    configure_env = cmeel_config.get_configure_env()
    configure_args = cmeel_config.get_configure_args(
        conf,
        install,
        configure_args,
        configure_env,
        run_tests,
    )
    configure_cmd = ["cmake", "-S", source, "-B", str(build), *configure_args]
    LOG.debug("configure environment: %s", configure_env)
    LOG.debug("configure command: %s", configure_cmd)
    check_call(configure_cmd, env=configure_env)

    LOG.info("build")
    build_cmd = ["cmake", "--build", str(build), f"-j{cmeel_config.jobs}"]
    LOG.debug("build command: %s", build_cmd)
    check_call(build_cmd)

    launch_tests(True, run_tests and not run_tests_after_install, pyproject, build)

    LOG.info("install")
    install_cmd = ["cmake", "--build", str(build), "-t", "install"]
    LOG.debug("install command: %s", install_cmd)
    check_call(install_cmd)

    launch_tests(False, run_tests and run_tests_after_install, pyproject, build)

    LOG.info("fix relocatablization")
    # Replace absolute install path in generated .cmake files, if any.
    for cmake_file in install.rglob("*.cmake"):
        ff = install / f"{cmake_file.stem}.fix"
        with cmake_file.open("r") as fr, ff.open("w") as fw:
            fw.write(fr.read().replace(str(install), "${PACKAGE_PREFIX_DIR}"))
        cmake_file.unlink()
        ff.rename(cmake_file)

    LOG.info("create dist-info")

    dist_info = wheel_dir / f"{distribution}.dist-info"
    dist_info.mkdir(parents=True)

    LOG.info("create dist-info / METADATA")

    with (dist_info / "METADATA").open("w") as f:
        requires = pyproject["build-system"]["requires"]
        f.write("\n".join(metadata(conf, requires, dist_info)))

    LOG.info("create dist-info / top level")
    with (dist_info / "top_level.txt").open("w") as f:
        f.write("")

    LOG.info("create dist-info / WHEEL")
    with (dist_info / "WHEEL").open("w") as f:
        f.write(
            "\n".join(
                [
                    "Wheel-Version: 1.0",
                    f"Generator: cmeel {__version__}",
                    "Root-Is-Purelib: false",
                    f"Tag: {get_tag(pyproject)}",
                    "",
                ],
            ),
        )

    expose_bin(install, wheel_dir, distribution)

    ensure_relocatable(check_relocatable, install, prefix)

    if editable:
        LOG.info("Add .pth in wheel")
        with (wheel_dir / f"{distribution}.pth").open("w") as f:
            f.write(str((install / SITELIB).absolute()))
    elif fix_pkg_config:
        LOG.info("fix pkg-config files")
        for fc in install.glob("**/*.pc"):
            with fc.open() as f:
                pc_file = f.read()
            if str(install) in pc_file:
                rel = str(fc.parent.relative_to(install))
                fix = "/".join(["${pcfiledir}"] + [".." for _ in rel.split("/")])
                LOG.warning("fix pkg-config %s: replace %s by %s", fc, install, fix)
                with fc.open("w") as f:
                    f.write(pc_file.replace(str(install), fix))

    LOG.info("wheel pack")
    pack = check_output(
        [
            sys.executable,
            "-m",
            "wheel",
            "pack",
            "--build-number",
            str(build_number),
            "-d",
            wheel_directory,
            str(wheel_dir),
        ],
    ).decode()
    LOG.debug("wheel pack output: %s", pack)
    parse_pack = re.search("Repacking wheel as (.*\\.whl)\\.\\.\\.", pack)
    if parse_pack is None:
        err = (
            "'wheel pack' command did not provide the .whl name, "
            "or the parser must be updated. Please report the following output: \n"
        )
        raise FileNotFoundError(err + pack)
    name = Path(parse_pack.group(1)).name
    LOG.debug("returning '%s'", name)

    LOG.info("done")
    return name



"""Run cmeel as a python module."""

import argparse
import logging
import os
import pathlib
import sys

from . import __version__
from .docker import add_docker_arguments, docker_build
from .env import add_paths_arguments, get_paths
from .release import add_release_arguments, release

LOG = logging.getLogger("cmeel")


def parse_args() -> argparse.Namespace:
    """Check what the user want."""
    # Get current interpreter
    python = pathlib.Path(sys.executable)
    if str(python.parent) in os.environ.get("PATH", "").split(os.pathsep):
        # its path is in PATH: no need for absolute path
        python = pathlib.Path(python.name)

    parser = argparse.ArgumentParser(
        prog=f"{python} -m cmeel",
        description="cmeel helpers",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="increment verbosity level",
    )
    subparsers = parser.add_subparsers(
        title="subcommands",
        description="valid sub-commands",
        help="sub-command help",
    )

    add_paths_arguments(subparsers)
    add_docker_arguments(subparsers)
    add_release_arguments(subparsers)

    ver = subparsers.add_parser("version", help="print current cmeel version.")
    ver.set_defaults(cmd="version")

    args = parser.parse_args()

    if args.verbose == 0:
        level = os.environ.get("CMEEL_LOG_LEVEL", "WARNING")
    else:
        level = 30 - 10 * args.verbose
    logging.basicConfig(level=level)

    LOG.debug("parsed arguments: %s", args)

    if "cmd" in args:
        LOG.debug("running subcommand %s", args.cmd)
        return args

    parser.print_help()
    sys.exit(0)


def main():
    """Run helpers."""
    args = parse_args()
    if args.cmd == "docker":
        docker_build(**vars(args))
    elif args.cmd == "release":
        release(**vars(args))
    elif args.cmd == "version":
        print(f"This is cmeel version {__version__}")
    else:
        print(get_paths(**vars(args)))


if __name__ == "__main__":
    main()



"""Cmeel run."""

import os
import sys
from pathlib import Path
from subprocess import run

from .consts import CMEEL_PREFIX


def cmeel_run():
    """Wrap an executable inside cmeel prefix."""
    prefix = Path(__file__).parent.parent / CMEEL_PREFIX  # TODO: not uniq

    exe = Path(sys.argv[0]).name
    sys.argv[0] = prefix / "bin" / exe

    # TODO: RPATH would be better
    lib = f"{prefix}/lib"
    if "LD_LIBRARY_PATH" in os.environ:
        if lib not in os.environ["LD_LIBRARY_PATH"]:
            os.environ["LD_LIBRARY_PATH"] = f"{lib}:{os.environ['LD_LIBRARY_PATH']}"
    else:
        os.environ["LD_LIBRARY_PATH"] = lib

    exe = run(sys.argv, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr)
    sys.exit(exe.returncode)



"""Utilities."""

import logging
import os
import re
import sys
import warnings
from importlib.util import find_spec
from pathlib import Path
from subprocess import CalledProcessError, check_call, check_output, run

from .config import cmeel_config

LOG = logging.getLogger("cmeel.utils")

PATCH_IGNORE = [
    "hunk ignored",
    "hunks ignored",
    "Skipping patch.",
    "The next patch would delete",
]

EXECUTABLE = ["#!python", "from cmeel.run import cmeel_run", "cmeel_run()"]


class PatchError(CalledProcessError):
    """Exception raised when patch operation failed."""

    def __str__(self):
        """Render this error as a string."""
        if self.returncode and self.returncode < 0:
            return super().__str__()
        return (
            f"Command '{self.cmd}' exit status {self.returncode}\n"
            f"with output:\n{self.output}\n"
            f"and stderr:\n{self.stderr}\n"
        )


class NonRelocatableError(Exception):
    """Exception raised when absolute paths are in the final package."""

    pass


def dotget(data, key, default):
    """Get key in data or default."""
    for part in key.split("."):
        if part in data:
            data = data[part]
        else:
            return default
    return data


def deprecate_build_system(pyproject, key, default):
    """Cmeel up to v0.22 was using the "build-system" section of pyproject.toml.

    This function helps to deprecate that and move to "tool.cmeel".
    """
    if key in pyproject["build-system"]:
        default = pyproject["build-system"][key]
        warnings.warn(
            'Using the "build-system" section of pyproject.toml for cmeel '
            "configuration is deprecated since cmeel v0.23 and will be removed in v1.\n"
            f'Please move your "{key} = {default}" to the "tool.cmeel" section.',
            DeprecationWarning,
            stacklevel=2,
        )
    if "tool" in pyproject and "cmeel" in pyproject["tool"]:
        return pyproject["tool"]["cmeel"].get(key, default)
    return default


def normalize(name: str) -> str:
    """Normalize name.

    ref. https://packaging.python.org/en/latest/specifications/name-normalization
    """
    return re.sub(r"[-_.]+", "-", name).lower()


def log_pip():
    """Log output of pip freeze."""
    if LOG.getEffectiveLevel() <= logging.DEBUG:
        if find_spec("pip") is not None:
            LOG.debug("pip freeze:")
            deps = check_output([sys.executable, "-m", "pip", "freeze"], text=True)
            for dep in deps.strip().split("\n"):
                LOG.debug("  %s", dep)


def get_tag(pyproject) -> str:
    """Find the correct tag for the wheel."""
    try:
        from packaging.tags import sys_tags
    except ImportError as e:
        err = "You need the 'build' extra option to use this build module.\n"
        err += "For this you can install the 'cmeel[build]' package."
        raise ImportError(err) from e

    tag = str(next(sys_tags()))
    # handle cross compilation on macOS with cibuildwheel
    # ref. https://github.com/pypa/cibuildwheel/blob/6549a9/cibuildwheel/macos.py#L221
    if "_PYTHON_HOST_PLATFORM" in os.environ:
        plat = os.environ["_PYTHON_HOST_PLATFORM"].replace("-", "_").replace(".", "_")
        tag = "-".join(tag.split("-")[:-1] + [plat])

    if deprecate_build_system(pyproject, "py3-none", False):
        warnings.warn(
            "The 'py3-none = true' key is deprecated. Please use 'has-sitelib = false'",
            DeprecationWarning,
            stacklevel=2,
        )
        tag = "-".join(["py3", "none", tag.split("-")[-1]])
    elif deprecate_build_system(pyproject, "any", False):
        warnings.warn(
            "The 'any = true' key is deprecated. "
            "Please use 'has-sitelib = false' and 'has-binaries = false'",
            DeprecationWarning,
            stacklevel=2,
        )
        tag = "py3-none-any"
    elif deprecate_build_system(pyproject, "pyver-any", False):
        warnings.warn(
            "The 'pyver-any = true' key is deprecated. "
            "Please use 'has-binaries = false'",
            DeprecationWarning,
            stacklevel=2,
        )
        tag = f"py3{sys.version_info.minor}-none-any"
    else:
        binaries = dotget(pyproject, "tool.cmeel.has-binaries", True)
        sitelib = dotget(pyproject, "tool.cmeel.has-sitelib", True)
        if not binaries and not sitelib:
            tag = "py3-none-any"
        elif not binaries:
            tag = f"py3{sys.version_info.minor}-none-any"
        elif not sitelib:
            tag = "-".join(["py3", "none", tag.split("-")[-1]])
    return tag


def patch():
    """Apply cmeel.patch if it exists and was not already applied."""
    if Path("cmeel.patch").exists():
        LOG.info("patching")
        cmd = ["patch", "-p0", "-s", "-N", "-i", "cmeel.patch"]
        ret = run(cmd, capture_output=True, text=True)
        if ret.returncode != 0:
            # If this patch was already applied, it's okay.
            for line in ret.stdout.split("\n"):
                if not line or any(val in line for val in PATCH_IGNORE):
                    continue
                raise PatchError(
                    returncode=ret.returncode,
                    cmd=cmd,
                    output=ret.stdout,
                    stderr=ret.stderr + f"\nwrong line: {line}\n",
                )
            LOG.info("this patch was already applied")


def expose_bin(install: Path, wheel_dir: Path, distribution: str):
    """Add scripts wrapping calls to CMEEL_PREFIX/bin/ executables."""
    bin_dir = install / "bin"
    if bin_dir.is_dir():
        LOG.info("adding executables")
        scripts = wheel_dir / f"{distribution}.data" / "scripts"
        scripts.mkdir(parents=True)
        for fn in bin_dir.glob("*"):
            executable = scripts / fn.name
            with fn.open("rb") as fo:
                is_script = fo.read(2) == b"#!"
            with executable.open("w") as fe:
                content = fn.read_text().split("\n") if is_script else EXECUTABLE
                if "python" in content[0]:
                    content = ["#!python", *content[1:]]
                fe.write("\n".join(content))
            executable.chmod(0o755)


def ensure_relocatable(check_relocatable: bool, install: Path, prefix: Path):
    """Ensure no cmake file contains wrong absolute paths."""
    if not check_relocatable:
        return
    LOG.info("check generated cmake files")
    wrong_dirs = [
        "/tmp/pip-build-env",
        "/tmp/pip-req-build",
        "/opt/_internal",
        str(prefix),
    ]
    for fc in install.glob("**/*.cmake"):
        with fc.open() as f:
            cmake_file = f.read()
            if any(wrong_dir in cmake_file for wrong_dir in wrong_dirs):
                lines = cmake_file.split("\n")
                # Get indexes of of problematic lines
                indexes = [
                    idx
                    for idx, line in enumerate(lines)
                    if any(wrong_dir in line for wrong_dir in wrong_dirs)
                ]
                # Get lines at those indexes and around them to display
                display = [
                    f"{i}: {line}"
                    for i, line in enumerate(lines)
                    if any(idx in indexes for idx in (i - 2, i - 1, i, i + 1, i + 2))
                ]
                raise NonRelocatableError(
                    f"{fc} references temporary paths:\n" + "\n".join(display),
                )


def launch_tests(before: bool, now: bool, pyproject, build: Path):
    """Launch tests, before or after the install."""
    if not now:
        return

    test_cmd = deprecate_build_system(
        pyproject,
        "test-cmd",
        ["cmake", "--build", "BUILD_DIR", "-t", "test"],
    )
    LOG.info("test {} install".format("before") if before else "after")
    test_env = cmeel_config.get_test_env()
    cmd = [i.replace("BUILD_DIR", str(build)) for i in test_cmd]
    LOG.debug("test environment: %s", test_env)
    LOG.debug("test command: %s", cmd)
    check_call(cmd, env=test_env)



"""Metadata generation from pyproject conf.

ref. PEP 621, superseeded by
https://packaging.python.org/en/latest/specifications/declaring-project-metadata/
"""

import warnings
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

LICENSE_GLOBS = ["LICEN[CS]E*", "COPYING*", "NOTICE*", "AUTHORS*"]


def get_license(conf: Dict[str, Any], dist_info: Optional[Path]) -> List[str]:
    """Parse 'license' and 'license-files' keys."""
    metadata = []

    lic_expr, lic_files = _license(conf)

    if "license-files" in conf:
        lic_files = [*lic_files, *_license_files(conf["license-files"])]
    elif not lic_files:
        for glob_expr in LICENSE_GLOBS:
            for lic_file_s in Path().glob(glob_expr):
                lic_files.append(str(lic_file_s))

    if not lic_expr and not lic_files:
        e = "'license' or 'license-files' is required"
        raise KeyError(e)

    if lic_expr:
        metadata.append(f"License-Expression: {lic_expr}")
    if dist_info:
        for lic_file in lic_files:
            metadata.append(f"License-File: {lic_file}")
            path_src = Path(lic_file)
            path_dst = dist_info / "licenses" / path_src
            path_dst.parent.mkdir(parents=True, exist_ok=True)
            with path_src.open("r") as f_src, path_dst.open("w") as f_dst:
                f_dst.write(f_src.read())

    return metadata


def _license_files(license_files: Union[str, List[str], Dict[str, str]]) -> List[str]:
    """Parse 'license-files' key."""
    lic_files = []
    if isinstance(license_files, str):
        lic_files.append(license_files)
    elif isinstance(license_files, list):
        lic_files += license_files
    elif isinstance(license_files, dict):
        if "paths" in license_files and "globs" not in license_files:
            for lic_file in license_files["paths"]:
                lic_files.append(lic_file)
        elif "paths" not in license_files and "globs" in license_files:
            for glob_expr in license_files["globs"]:
                for lic_file_s in Path().glob(glob_expr):
                    lic_files.append(str(lic_file_s))
        else:
            e = "'license-files' table must containe either a 'paths' or a 'globs'"
            raise KeyError(e)
    else:
        e = "'license-files' accepts either a string, a list, or a table."
        raise TypeError(e)

    return lic_files


def _license(conf) -> Tuple[str, List[str]]:
    """Parse 'license' key."""
    lic_expr, lic_files = "", []
    if "license" in conf:
        if isinstance(conf["license"], str):
            lic_expr = conf["license"]
        elif isinstance(conf["license"], dict):
            warnings.warn(
                "'license' table is deprecated.\n"
                "Please use a 'license' string and/or the 'license-files' key.\n"
                f"The default setting globs {LICENSE_GLOBS}, as per PEP 639",
                DeprecationWarning,
                stacklevel=2,
            )
            if "text" in conf["license"] and "file" not in conf["license"]:
                lic_expr = conf["license"]["text"]
            elif "text" not in conf["license"] and "file" in conf["license"]:
                lic_files.append(conf["license"]["file"])
            else:
                e = "'license' table must containe either a 'file' or a 'text'"
                raise KeyError(e)
        else:
            e = "'license' accepts either a string or a table."
            raise TypeError(e)
    return lic_expr, lic_files


def get_people(conf: Dict[str, Any], key: str) -> List[str]:
    """Parse 'authors' and 'maintainers' keys."""
    metadata = []

    names, mails = [], []

    for person in conf.get(f"{key}s", []):
        if "name" in person and "email" in person:
            mails.append(f"{person['name']} <{person['email']}>")
        elif "email" in person:
            mails.append(person["email"])
        elif "name" in person:
            names.append(person["name"])

    if names:
        metadata.append(f"{key.title()}: " + ",".join(names))
    if mails:
        metadata.append(f"{key.title()}-email: " + ",".join(mails))

    return metadata


def get_urls(conf: Dict[str, Any]) -> List[str]:
    """Parse 'urls' keys."""
    metadata = []

    if "urls" in conf:
        for key, url in conf["urls"].items():
            if key == "homepage":
                metadata.append(f"Home-page: {url}")
            else:
                name = key.replace("-", " ").capitalize()
                metadata.append(f"Project-URL: {name}, {url}")

    return metadata


def get_deps(conf: Dict[str, Any], build_deps: List[str]) -> List[str]:
    """Parse 'dependencies' keys."""
    metadata = []

    dependencies = ["cmeel", *conf.get("dependencies", [])]
    for dep in dependencies:
        metadata.append(f"Requires-Dist: {dep}")

    build_dependencies = [
        build_dep
        for build_dep in build_deps
        if build_dep != "cmeel[build]" and build_dep not in dependencies
    ]
    if build_dependencies:
        metadata.append("Provides-Extra: build")
        for build_dep in build_dependencies:
            metadata.append(f'Requires-Dist: {build_dep} ; extra == "build"')

    for extra, deps in conf.get("optional-dependencies", []):
        if extra == "build":
            e = "the 'build' extra is reserved by cmeel."
            raise ValueError(e)
        metadata.append(f"Provides-Extra: {extra}")
        for dep in deps:
            metadata.append(f'Requires-Dist: {dep} ; extra == "{extra}"')

    return metadata


def get_readme(conf: Dict[str, Any]) -> List[str]:
    """Parse 'readme' key."""
    metadata = []

    readme_file, readme_content, readme_type = "", "", ""
    if "readme" not in conf:
        for ext in [".md", ".rst", ".txt", ""]:
            if Path(f"README{ext}").exists():
                conf["readme"] = f"README{ext}"
                break
    if "readme" in conf:
        if isinstance(conf["readme"], str):
            readme_file = conf["readme"]
            readme_type = _ext_type(conf["readme"])
        elif isinstance(conf["readme"], dict):
            readme_file, readme_content, readme_type = _readme_dict(conf)
        else:
            e = "'readme' accepts either a string or a table."
            raise TypeError(e)
        metadata.append(f"Description-Content-Type: {readme_type}")

        metadata.append("")

        if readme_content:
            metadata.append(readme_content)
        elif readme_file:
            with Path(readme_file).open() as f:
                metadata.append(f.read())

    return metadata


def _readme_dict(conf: Dict[str, Any]) -> Tuple[str, str, str]:
    """Parse 'readme' key when it is a table."""
    readme_file, readme_content, readme_type = "", "", ""

    if "content-type" in conf["readme"]:
        readme_type = conf["readme"]["content-type"]
    else:
        e = "if 'readme' is a table, it must contain a 'content-type' key"
        raise KeyError(e)
    if "file" in conf["readme"] and "text" not in conf["readme"]:
        readme_file = conf["readme"]["file"]
    elif "file" not in conf["readme"] and "text" in conf["readme"]:
        readme_content = conf["readme"]["text"]
    else:
        e = "'readme' table accepts either a 'file' or a 'text' key"
        raise KeyError(e)

    return readme_file, readme_content, readme_type


def _ext_type(filename: str) -> str:
    """Guess a content type based on extension."""
    if filename.lower().endswith(".md"):
        return "text/markdown"
    if filename.lower().endswith(".rst"):
        return "text/x-rst"
    return "text/plain"


def get_keywords(conf: Dict[str, Any]) -> List[str]:
    """Parse 'keyword' key."""
    metadata = []
    if "keywords" in conf:
        keywords = ",".join(conf["keywords"])
        metadata.append(f"Keywords: {keywords}")
    return metadata


def metadata(conf, requires: List[str], dist_info: Optional[Path] = None) -> List[str]:
    """Return the lines which should go in the METADATA / PKG-INFO file."""
    return [
        "Metadata-Version: 2.4",
        f"Name: {conf['name']}",
        f"Version: {conf['version']}",
        f"Summary: {conf['description']}",
        f"Requires-Python: {conf.get('requires-python', '>=3.8')}",
        *get_license(conf, dist_info),
        *get_people(conf, "author"),
        *get_people(conf, "maintainer"),
        *get_keywords(conf),
        *get_urls(conf),
        *get_deps(conf, requires),
        *[f"Classifier: {classifier}" for classifier in conf.get("classifiers", [])],
        *get_readme(conf),
    ]



"""PEP 517 & 660 entry points."""

import logging
import os

from .impl import build_impl
from .sdist import sdist_impl

LOG = logging.getLogger("cmeel")


def build_editable(
    wheel_directory,
    config_settings=None,
    metadata_directory=None,
) -> str:
    """Build an editable wheel: main entry point for PEP 660."""
    LOG.info("cmeel build editable")
    os.environ["CMAKE_INSTALL_MODE"] = "ABS_SYMLINK"
    return build_impl(wheel_directory, editable=True)


def build_wheel(wheel_directory, config_settings=None, metadata_directory=None) -> str:
    """Build a binary wheel: main entry point for PEP 517."""
    LOG.info("cmeel build wheel")
    return build_impl(wheel_directory, editable=False)


def build_sdist(sdist_directory, config_settings=None) -> str:
    """Generate a gzipped distribution tarball."""
    return sdist_impl(sdist_directory)



"""Build a project with cmeel in a container."""

import logging
import os
import pathlib
from subprocess import check_call
from typing import List, Optional

from .backports import BooleanOptionalAction

LOG = logging.getLogger("cmeel.docker")


def add_docker_arguments(subparsers):
    """Append docker command for argparse."""
    sub = subparsers.add_parser("docker", help="build a project in a container.")
    sub.add_argument(
        "-i",
        "--image",
        default="quay.io/pypa/manylinux_2_28_x86_64",
        help="docker image to use for building the wheel",
    )
    sub.add_argument(
        "-p",
        "--python",
        default="python3.13",
        help="python interpreter inside that image",
    )
    sub.add_argument(
        "-u",
        "--update",
        action="store_true",
        help="update docker image",
    )
    sub.add_argument(
        "-U",
        "--upgrade",
        action="store_true",
        help="upgrade pip",
    )
    sub.add_argument(
        "-c",
        "--cache",
        action="store_true",
        help="binds /root/.cache/pip",
    )
    sub.add_argument(
        "-C",
        "--cwd",
        default=str(pathlib.Path.cwd()),
        help="build the project in this directory",
    )
    sub.add_argument(
        "-e",
        "--env",
        action="append",
        help="pass environment variables to docker run",
    )
    sub.add_argument(
        "--cmeel-env",
        action=BooleanOptionalAction,
        default=True,
        help="forward 'CMEEL_*' environment variables to docker run",
    )
    sub.set_defaults(cmd="docker")


def docker_build(
    image: str,
    python: str,
    update: bool,
    cache: bool,
    upgrade: bool,
    cwd: str,
    env: Optional[List[str]],
    cmeel_env: bool,
    **kwargs,
):
    """Build a project with cmeel in a container."""
    if update:
        pull = ["docker", "pull", image]
        LOG.info("running '%s'", pull)
        check_call(pull)

    volumes = ["-v", f"{cwd}/:/src"]
    envs: List[str] = []
    if env:
        for e in env:
            envs = [*envs, "-e", e]
    if cmeel_env:
        for e in os.environ:
            if e.startswith("CMEEL_"):
                envs = [*envs, "-e", e]
    if cache:
        volumes = [*volumes, "-v", "/root/.cache/pip:/root/.cache/pip"]
    docker = ["docker", "run", "--rm", *envs, *volumes, "-w", "/src", "-t", image]
    build = [python, "-m", "pip", "wheel", "-vw", "wh", "."]
    if upgrade:
        pip = [python, "-m", "pip", "install", "-U", "pip"]
        pip_cmd = " ".join(pip)
        build_cmd = " ".join(build)
        cmd = [*docker, "bash", "-c", f"{pip_cmd} && {build_cmd}"]
    else:
        cmd = [*docker, *build]
    LOG.info("running '%s'", cmd)
    check_call(cmd)



"""Tools to help environment management."""

import os
import pathlib
import sys

from .consts import CMEEL_PREFIX

PATHS = {
    "cmake": "CMAKE_PREFIX_PATH",
    "lib": "LD_LIBRARY_PATH",
    "pc": "PKG_CONFIG_PATH",
}


def add_paths_arguments(subparsers):
    """Append paths commands for argparse."""
    for cmd, path in PATHS.items():
        sub = subparsers.add_parser(cmd, help=f"show cmeel additions to {path}")
        sub.add_argument("--prepend", action="store_true", help=f"show full {path}")
        sub.set_defaults(cmd=cmd)


def get_paths(cmd: str, prepend: bool = False, **kwargs) -> str:
    """Get the paths needed by the user."""
    prefixes = [pathlib.Path(path) / CMEEL_PREFIX for path in sys.path]
    if cmd == "lib":
        prefixes = [p / "lib" for p in prefixes]
    elif cmd == "pc":
        prefixes = [p / sub / "pkgconfig" for p in prefixes for sub in ["lib", "share"]]

    available = [str(p) for p in prefixes if p.is_dir()]
    if prepend:
        ret = []
        for prefix in available + os.environ.get(PATHS[cmd], "").split(os.pathsep):
            if prefix and prefix not in ret:
                ret.append(prefix)
        return os.pathsep.join(ret)
    return os.pathsep.join(available)



#!/usr/bin/env python
"""Helper to release a cmeel project."""

from logging import getLogger
from pathlib import Path
from subprocess import check_call

try:
    import tomllib  # type: ignore
except ModuleNotFoundError:
    import tomli as tomllib  # type: ignore

LOG = getLogger("cmeel.release")


def add_release_arguments(subparsers):
    """Append release command for argparse."""
    sub = subparsers.add_parser("release", help="release a cmeel project.")
    sub.set_defaults(cmd="release")


def release(**kwargs):
    """Release a cmeel project."""
    with Path("pyproject.toml").open("rb") as f:
        pyproject = tomllib.load(f)

    version = pyproject["project"]["version"]
    build = pyproject.get("tool", {}).get("cmeel", {}).get("build-number", 0)
    tag = f"v{version}.c{build}"
    release = f"Cmeel Release {tag}"

    LOG.info("Releasing vesion '%s'", tag)

    commit_cmd = ["git", "commit", "-am", release]
    LOG.debug("commit command: %s", commit_cmd)
    check_call(commit_cmd)

    tag_cmd = ["git", "tag", "-s", tag, "-m", release]
    LOG.debug("tag command: %s", tag_cmd)
    check_call(tag_cmd)

    push_cmd = ["git", "push", "origin", tag]
    LOG.debug("push command: %s", push_cmd)
    check_call(push_cmd)


if __name__ == "__main__":
    import logging

    logging.basicConfig(level=logging.DEBUG)
    release()



"""Generate .tar.gz source distribution."""

import logging
import tarfile
from pathlib import Path
from tempfile import TemporaryDirectory

try:
    import tomllib  # type: ignore
except ModuleNotFoundError:
    import tomli as tomllib  # type: ignore

from .metadata import metadata
from .utils import normalize

LOG = logging.getLogger("cmeel.sdist")


def sdist_impl(sdist_directory) -> str:
    """Implement the build_sdist entry point."""
    try:
        import git_archive_all
    except ImportError as e:
        err = "You need the 'build' extra option to use this build module.\n"
        err += "For this you can install the 'cmeel[build]' package."
        raise ImportError(err) from e

    LOG.info("load conf from pyproject.toml")
    with Path("pyproject.toml").open("rb") as f:
        pyproject = tomllib.load(f)

    conf = pyproject["project"]
    conf["name"] = normalize(conf["name"])
    distribution = f"{conf['name'].replace('-', '_')}-{conf['version']}"

    # tarfile can't add PKG-INFO to a .tar.gz, so we have to make a tmp one
    with TemporaryDirectory() as tmp:
        tmp_pkg = Path(tmp) / "PKG-INFO"
        tmp_tar = Path(tmp) / f"{distribution}.tar.gz"
        def_tar = Path(sdist_directory) / f"{distribution}.tar.gz"

        LOG.info("archive git repository and its submodules in {tmp}")
        git_archive_all.main(
            ["git_archive_all.py", str(tmp_tar)],
        )

        LOG.info("write PKG-INFO file")
        requires = pyproject["build-system"]["requires"]
        with tmp_pkg.open("w") as f:
            f.write("\n".join(metadata(conf, requires)))

        LOG.info("create final archive with previous one + PKG-INFO")
        with tarfile.open(tmp_tar, "r") as tr, tarfile.open(def_tar, "w:gz") as tw:
            for member in tr.getmembers():
                fileobj = None
                if member.type not in [tarfile.LNKTYPE, tarfile.SYMTYPE]:
                    fileobj = tr.extractfile(member)
                tw.addfile(member, fileobj)
            tw.add(str(tmp_pkg), f"{distribution}/PKG-INFO")

    return distribution
